(* $Novell: extmatch.c,v 1.10 2003/05/12 12:19:53 $ */
/**************************************************************************
*  Novell Software Developer Kit
*
*  Copyright (C) 2002-2003 Novell, Inc. All Rights Reserved.
*
*  THIS WORK IS SUBJECT TO U.S. AND INTERNATIONAL COPYRIGHT LAWS AND TREATIES.
*  USE AND REDISTRIBUTION OF THIS WORK IS SUBJECT TO THE LICENSE AGREEMENT
*  ACCOMPANYING THE SOFTWARE DEVELOPER KIT (SDK) THAT CONTAINS THIS WORK.
*  PURSUANT TO THE SDK LICENSE AGREEMENT, NOVELL HEREBY GRANTS TO DEVELOPER A
*  ROYALTY-FREE, NON-EXCLUSIVE LICENSE TO INCLUDE NOVELL'S SAMPLE CODE IN ITS
*  PRODUCT. NOVELL GRANTS DEVELOPER WORLDWIDE DISTRIBUTION RIGHTS TO MARKET,
*  DISTRIBUTE, OR SELL NOVELL'S SAMPLE CODE AS A COMPONENT OF DEVELOPER'S
*  PRODUCTS. NOVELL SHALL HAVE NO OBLIGATIONS TO DEVELOPER OR DEVELOPER'S
*  CUSTOMERS WITH RESPECT TO THIS CODE.
*
***************************************************************************
   extMatch.c 
***************************************************************************
   Description: The extMatch.c sample constructs the extensible match 
                filter with the given inputs,  searches the directory and 
                prints the results.  Extensible match requires eDirectory
                version 8.7 or higher.
   
   The LDAP v3 Core protocol specification requires LDAP servers to 
   recognize a search element called an extensible match. Extensible
   match provides additional features which can be expressed in the
   search filter.
   
   Feature 1. Allow matching to components of the object's Distinguished Name 
              as well as attributes of the object.
        
   Feature 2. Allow a specific matching rule to be used for an attribute-value 
              comparison, such as a sounds-alike match.
        
   Feature 3. Allow a match to any attribute of an object supporting a 
              particular matching rule.  For example, one could search for 
              objects with any string attribute containing "Dino".
   
   These features may be used independently or in combination.

   All the supported matching rules are advertised through LDAP subschema
   subentry object using the standard matchingRules and matchingRuleUse
   schema attributes.

   It is mandatory that either dnAttribute or matching rule be present
   as part of extensible match filter. 
   
   NOTE:  In eDirectory 8.7, only the DN feature of extensible match 
   is supported (feature 1).  
   
   The following examples illustrate the use of extensible match filters
   and correspond to the features listed above.

   Example #1: (ou:dn:=Sales)
   Same as filter "(ou=Sales)" except that DN components should be 
   considered part of the entry when doing the match.  Any object 
   with "ou=Sales" as part of its DN will match.

   Example #2: (cn:2.4.6.8:=Barney Rubble)
   Specifies matching rule "2.4.6.8" to be used when making comparisons
   for (cn="Barney Rubble").

   Example #2a: (cn:dn:2.4.6.8:=Barney Rubble)
   Combines examples #1 and #2.  Specifies the matching rule to use and 
   indicates that components of entry's DN should be considered attributes of
   the entry when evaluating the match.

   Example #3: (:2.4.6.8:=Dino)
   Indicates this filter should be applied to all attributes supporting
   this matching rule.

   Example #3a: (:dn:2.4.6.8:=Dino)
   Same as example #3, but also includes components of the DN in the match.

   For further details, refer to IETF RFC 2251 and RFC 2254.

***************************************************************************)
program extmatch;

uses sysutils, ldap;

procedure usage;
begin
  writeln('extMatch [-d] <host name> <port number> <login dn> <password>');
  writeln('         <search base> <attribute name> <matching rule> <match value>');
  writeln('where:');
  writeln('  -d means consider DN components as attributes of the object');
  writeln('    <search base> is the starting container for a subtree search');
  writeln('    <attribute name> is the attribute to match against, or '''' for all attrs');
  writeln('    <matching rule> is an OID specifying the matching rule to use, or ''''');
  writeln('    <match value> is the value to compare against in the search filter');
  writeln('Examples:');
  writeln(' #1. extMatch -d Acme.com 389 cn=admin,o=Acme secret o=Acme ou '''' Sales');
  writeln(' #2. extMatch Acme.com 389 cn=admin,o=Acme secret');
  writeln('              ou=Sales,o=Acme cn 2.4.6.8 ''Barney Rubble''');
  writeln(' #3. extMatch Acme.com 389 cn=admin,o=Acme secret o=Acme '''' 2.4.6.8 Dino');
  writeln('              NOTE: Only example 1 is supported by eDirectory 8.7');
end;

function remove_delim(const str: AnsiString): AnsiString;
begin
  result := str;
  if (length(str)>=2) then begin
    if (str[1]='''') and (str[length(str)]='''') then begin
      result := Copy(Str,2,length(str)-2);
    end;
  end;
end;

const
    timeout : timeval = (tv_sec:10;tv_usec:0); //* 10 second connection timeout */
var
    version, ldapPort, rc: int32;
    ldapHost, loginDN, password, searchBase : AnsiString;
    dn: pchar8;
    matchAttribute, matchingRule, matchValue, extMatchFilter: AnsiString;
    ld:  hLDAP;
    searchResult, entry: LDAPMessage;
    attrs: array[0..1] of pchar8;
    useDN, useRule, useAttribute: boolean;

    argBase: integer;
begin
    attrs[0] := LDAP_NO_ATTRS;  //* Return DN names only */
    attrs[1] := nil;
    useDN := false;                 //* 1 if to use dnAttribute */
    useRule := false;               //* 1 if matching rule is given */
    useAttribute := false;          //* 1 if attribute value is given */

    argBase := 0;

    //* Determine which options are set */
    if (ParamCount()>0) then begin
       //* -d option inserts ":dn" in the filter */
       if (ParamStr(1)='-d') then begin
          useDN := true;
          argBase := 1;
       end else begin;
          //* An option, but not -d option */
          if (Copy(ParamStr(1),1,1)='-') then begin
             usage;
             halt(1);
          end;
       end;
    end;

    //* Validate number of arguments against options chosen. */
    if (ParamCount()=(8+argBase)) then begin
        ldapHost        := ParamStr(1+argBase);
        ldapPort        := StrToInt(ParamStr(2+argBase));
        loginDN         := ParamStr(3+argBase);
        password        := ParamStr(4+argBase);
        searchBase      := ParamStr(5+argBase);
        matchAttribute  := remove_delim(ParamStr(6+argBase));
        matchingRule    := remove_delim(ParamStr(7+argBase));
        matchValue      := remove_delim(ParamStr(8+argBase));
    end else begin
        usage;
        halt(1);
    end;

    if (matchAttribute<>'') then begin
       useAttribute := true;
    end;   

    if (matchingRule<>'') then begin
       useRule := true;
    end;

    //* Must have dn option or matching rule. */
    if (not useAttribute) and (not useRule) then begin
        writeln('Error: Either <attribute name> or <matching rule> must be '+
                 'present as part of an extensible match filter.');
        usage;
        halt(1);
    end;
    
    //* Construct extensible matching filter - "(attr:dn:rule:=value)" */
    extMatchFilter := '(';
    if (useAttribute) then begin
      extMatchFilter := extMatchFilter + matchAttribute;
    end;
    if (useDN) then begin
      extMatchFilter := extMatchFilter + ':dn';
    end;
    if (useRule) then begin
      extMatchFilter := extMatchFilter + ':' + matchingRule;
    end;
    extMatchFilter := extMatchFilter + ':=' + matchValue + ')';

    writeln('    Extensible match filter:  ', extMatchFilter);

    //* Set LDAP version to 3 and set connection timeout. */
    version := LDAP_VERSION3;
    writeln('ldap_set_option');
    ldap_set_option(NIL, LDAP_OPT_PROTOCOL_VERSION, version);
    ldap_set_option(NIL, LDAP_OPT_NETWORK_TIMEOUT, timeOut);

    //* Initialize the LDAP session */
    writeln('ldap_init');
    ld := ldap_init(pchar8(ldapHost),ldapPort);

    if (ld = nil) then begin
      writeln('LDAP session initialization failed');
      halt(1);
    end;
    writeln('LDAP session initialized');

    //* Bind to the server */
    writeln('ldap_simple_bind_s');
    rc := ldap_simple_bind_s(ld, pchar8(loginDN), pchar8(password));
    if (rc <> LDAP_SUCCESS) then begin
        writeln('ldap_simple_bind_s: ', ldap_err2string(rc));
        ldap_unbind_s(ld);
        halt(1);
    end;
    writeln('Bind successful');

    //* Search the directory */
    rc := ldap_search_ext_s(
                    ld,                    //* LDAP session handle */
                    pchar8(searchBase),    //* container to search */
                    LDAP_SCOPE_SUBTREE,    //* search scope */
                    pchar8(extMatchFilter),//* extensible match search filter */
                    addr(attrs),           //* attributes to return */
                    0,                     //* return attributes and values */
                    nil,                   //* server controls */
                    nil,                   //* client controls */
                    addr(timeOut),         //* search timeout */
                    LDAP_NO_LIMIT,         //* no size limit */
                    searchResult);         //* returned results */

    if (rc <> LDAP_SUCCESS) then begin
        writeln('ldap_search_ext_s: ', ldap_err2string(rc));
        ldap_msgfree(searchResult);
        ldap_unbind_s(ld);
        halt(1);
    end;

    //* Read and print matched entry names. */
    entry :=  ldap_first_entry(ld, searchResult);
    while (entry<>nil) do begin
        dn := ldap_get_dn(ld,entry);
        if (dn<>nil) then begin
            writeln('  dn: ', dn);
            ldap_memfree(dn);
        end;
        entry := ldap_next_entry(ld, entry);
    end{while};

    writeln('Extensible match search completed successfully.');

    ldap_msgfree(searchResult);

    ldap_unbind_s( ld );
end.
